local ipairs = ipairs
local print = print
local select = select
local strsplit = strsplit
local strtrim = strtrim
local string = string
local table = table
local _GetTime = GetTime

local loaded = false
local quantity = 0
local lastKill = {0} -- To be populated later, do not remove the initial value. The zero means inconclusive/invalid data.
local nameplateUpdateTimer = 0
local activeNameplates = {}

mppDebug = false

mppSimulationMode = false
local simulationMax = 1000
local simulationCurrent = 350

local version = "0.1"
local defaultSettings = {
	version = version,
	enabled = true,
	inconclusiveDataThreshold = 100, -- Mobs killed within this span of time (in milliseconds) will not be processed since we might not get the criteria update fast enough to know which mob gave what progress. Well, that's the theory anyway.
	maxTimeSinceKill = 600, -- Lag tolerance between a mob dying and the progress criteria updating, in milliseconds.
	tooltipColor = "82E0FF", -- Color of the tooltip. Duh.
	nameplateUpdateRate = 250, -- Rate (in milliseconds) at which we update the progress we get from the current pull, as estimated by active name plates you're in combat with.
	enablePullEstimate = true,
	enableTooltip = true,
}

mppDefaultValues = {{97119,1,"Shroud Hound"},{96934,2,"Valarjar Trapper"},{102104,4,"Enslaved Shieldmaiden"},{98900,4,"Wyrmtongue Trickster"},{98919,4,"Seacursed Swiftblade"},{99033,4,"Helarjar Mistcaller"},{106786,1,"Bitterbrine Slave"},{91796,10,"Skrog Wavecrasher"},{95832,2,"Valarjar Shieldmaiden"},{91792,10,"Stormwake Hydra"},{91781,4,"Hatecoil Warrior"},{91785,2,"Wandering Shellback"},{97171,10,"Hatecoil Arcanist"},{96608,2,"Ebonclaw Worg"},{95947,1,"Mak'rana Hardshell"},{106785,1,"Bitterbrine Slave"},{100526,1,"Tormented Bloodseeker"},{91793,1,"Seaspray Crab"},{99365,4,"Taintheart Stalker"},{95939,10,"Skrog Tidestomper"},{101991,4,"Nightmare Dweller"},{96640,2,"Valarjar Marksman"},{106787,1,"Bitterbrine Slave"},{98810,6,"Wrathguard Bladelord"},{102094,4,"Risen Swordsman"},{97097,4,"Helarjar Champion"},{100531,8,"Bloodtainted Fury"},{96664,2,"Valarjar Runecarver"},{99188,4,"Waterlogged Soul Guard"},{97365,4,"Seacursed Mistmender"},{95920,2,"Animated Storm"},{95861,4,"Hatecoil Oracle"},{100527,3,"Dreadfire Imp"},{100216,4,"Hatecoil Wrangler"},{91786,1,"Gritslime Snail"},{97200,4,"Seacursed Soulkeeper"},{96574,1,"Stormforged Sentinel"},{91790,4,"Mak'rana Siltwalker"},{91794,1,"Saltscale Lurker"},{99358,4,"Rotheart Dryad"},{95834,2,"Valarjar Mystic"},{99366,4,"Taintheart Summoner"},{95842,2,"Valarjar Thundercaller"},{99307,8,"Skjal"},{97173,4,"Restless Tides"},{97182,6,"Night Watch Mariner"},{102788,4,"Felspite Dominator"},{97185,10,"The Grimewalker"},{91782,10,"Hatecoil Crusher"},{101414,2,"Saltscale Skulker"},{97197,2,"Valarjar Purifier"},{91783,4,"Hatecoil Stormweaver"},{97172,1,"Saltsea Droplet"},{98973,1,"Skeletal Warrior"},{97087,2,"Valarjar Champion"},{98792,4,"Wyrmtongue Scavenger"},{102375,3,"Runecarver Slave"},{98813,4,"Bloodscent Felhound"},{99360,9,"Vilethorn Blossom"},{100529,1,"Hatespawn Slime"},} -- Yeah ok look I know this is far from optimal, but I'm tired and this was a super quick way of just getting it in there.

-- New to lua, don't judge too hard pls.

function mppGetLastKill()
	return lastKill
end

--
-- GENERAL ADDON UTILITY
-- And by "utility" I mostly mean creating a bunch of shit that should really be built-in.

local function debugLog(s)
	if mppDebug then
		print(s)
	end
end

local function split(str)
	a = {}
	for s in string.gmatch(str, "%S+") do
		table.insert(a, s)
	end
	return a
end

local function round(number, decimals)
    return (("%%.%df"):format(decimals)):format(number)
end

local function tlen(t)
	local length = 0
	for _ in pairs(t) do
		length = length + 1
	end
	return length
end

local function GetTime()
	return _GetTime() * 1000
end


local function getSetting(setting)
	if MythicPlusProgressDB["settings"][setting] == nil then
		print("MPP attempted to get missing setting: " .. setting)
		return
	end
	return MythicPlusProgressDB["settings"][setting]
end

function mppGetSetting(setting)
	return getSetting(setting)
end

local function setSetting(setting, value)
	if MythicPlusProgressDB["settings"][setting] == nil then
		print("MPP attempted to get missing setting: " .. setting)
		return
	end
	MythicPlusProgressDB["settings"][setting] = value
	return value
end

local function toggleSetting(setting)
	return setSetting(setting, not getSetting(setting))
end

--
-- WOW GENERAL WRAPPERS/EZUTILITIES
--

local function getNPCID(guid)
	local targetType, _,_,_,_, npcID = strsplit("-", guid)
	if targetType == "Creature" and npcID then
		return tonumber(npcID)
	end
end

-- TODO: Figure out how to filter out bosses.
local function isValidTarget(targetToken)
	if UnitCanAttack("player", targetToken) then
		return true
	end
end

local function getSteps()
	return select(3, C_Scenario.GetStepInfo())
end

local function isDungeonFinished()
	if mppSimulationMode then return false end
	return (getSteps() and getSteps() < 1)
end

-- Will also return true in challenge modes if those are ever re-implemented as M+ is basically recycled Challenge Mode.
local function isMythicPlus()
	if mppSimulationMode then return true end
	local difficulty = select(3, GetInstanceInfo()) or -1
	if difficulty == 8 and not isDungeonFinished() then
		return true
	else
		return false
	end
end

-- DEPRECATED
local function isTeeming()
	print("Deprecated function isTeeming called!")
	return
	--[[local affixes = select(2, C_ChallengeMode.GetActiveKeystoneInfo())
	if affixes then
		for k,v in pairs(affixes) do
			if v == 5 then
				return true
			end
		return false
		end
	end--]]
end

function getProgressInfo()
	if isMythicPlus() then
		local numSteps = select(3, C_Scenario.GetStepInfo()) -- GetStepInfo tells us how many steps there are.
		if numSteps and numSteps > 0 then
			local info = {C_Scenario.GetCriteriaInfo(numSteps)} -- It should be the last step.
			if info[13] == true then -- if isWeightedProgress
				return info
			end
		end
	end
end
	
local function getMaxQuantity()
	if mppSimulationMode then return simulationMax end
	return getProgressInfo()[5]
end

local function getCurrentQuantity()
	if mppSimulationMode then return simulationCurrent end
	return strtrim(getProgressInfo()[8], "%")
end

local function getEnemyForcesProgress()
--	debugLog("getEnemyForcesProgress called.")
	-- Returns exact float value of current enemies killed progress (1-100).
	local quantity, maxQuantity = getCurrentQuantity(), getMaxQuantity()
	local progress = quantity / maxQuantity
	return progress * 100
end

--
-- DB READ/WRITES
--

local function getValue(npcID)
	debugLog("getValue called. Args: " .. npcID)
	local npcData = MythicPlusProgressDB["npcData"][npcID]
	if npcData then
		local hiValue, hiOccurrence = nil, -1
		for value, occurrence in pairs(npcData["values"]) do
			if occurrence > hiOccurrence then
				hiValue, hiOccurrence = value, occurrence
			end
		end
		if hiValue ~= nil then
			return hiValue
		end
	end
end

local function updateValue(npcID, value, npcName, forceQuantity)
	debugLog("updateValue called. Args: " .. npcID .. ", " .. value .. ", " .. npcName .. ", " .. forceQuantity)
	local npcData = MythicPlusProgressDB["npcData"][npcID]
	if not npcData then
		MythicPlusProgressDB["npcData"][npcID] = {values={}, name=npcName or "Unknown"}
		return updateValue(npcID, value, npcName, forceQuantity)
	end
	values = npcData["values"]
	if values[value] == nil then
		values[value] = forceQuantity or 1
		return false
	else
		values[value] = values[value] + forceQuantity or 1
		return true
	end
end

-- Temp testing global access
function mppUpdateValue(npcID, value)
	return updateValue(npcID, value)
end

-- Temp testing global access
function mppGetValue(npcID, value)
	return getValue(npcID)
end

--
-- Light DB wrap
--

-- Returns a nil-100 number representing the percentual progress that npcID is expected to give you.
local function getEstimatedProgress(npcID)
	debugLog("getEstimatedProgress called. Args: " .. npcID)
	local npcValue, maxQuantity = getValue(npcID), getMaxQuantity()
	if npcValue and maxQuantity then
		return (npcValue / maxQuantity) * 100
	end
end

--
-- TRIGGERS/HOOKS
--

-- Called when our enemy forces criteria increases, no matter how small the increase (but >0).
local function onProgressUpdated(deltaProgress)
	debugLog("onProgressUpdated called. Args: " .. deltaProgress)
	if currentQuantity == getMaxQuantity() then
		return -- Disregard data that caps us as we don't know if we got the full value.
	end
	local timestamp, npcID, npcName, isDataUseful = unpack(lastKill) -- See what the last mob we killed was
	if timestamp and npcID and deltaProgress and isDataUseful then -- Assert that we have some useful data to work with
		local timeSinceKill = GetTime() - timestamp
		debugLog("timeSinceKill: " .. timestamp .. " Current Time: " .. GetTime() .. "Timestamp of kill: " .. timeSinceKill)
		if timeSinceKill <= getSetting("maxTimeSinceKill") then
			debugLog(string.format("Gained %f%s. Last mob killed was %s (%i) %fs ago", deltaProgress, "%", npcName, npcID, timeSinceKill/1000))
			local success = updateValue(npcID, deltaProgress, npcName) -- Looks like we have ourselves a valid entry. Set this in our database/list/whatever.
			debugLog("updateValue: " ..npcID.. " - " ..success)
		else
			debugLog(string.format("Gained %f%s. Last mob killed was %s (%i) %fs ago (PAST CUTOFF!)", deltaProgress, "%", npcName, npcID, timeSinceKill))
		end
	end
end

-- Called directly by our hook
local function onCriteriaUpdate()
	debugLog("onCriteriaUpdate called")
	if not currentQuantity then
		currentQuantity = 0
	end
	if not isMythicPlus() or not loaded or not getSetting("enabled") then return end
	newQuantity = getCurrentQuantity()
	deltaQuantity = newQuantity - currentQuantity
	--print(newQuantity)
	--print(deltaQuantity)
	--print(currentQuantity)
	if deltaQuantity > 0 then
		currentQuantity = newQuantity
		onProgressUpdated(deltaQuantity)
	end
		
end

-- Called directly by our hook
local function onCombatLogEvent(args)
	--local _,combatType,_,_,_,_,_, destGUID, destName = unpack(args)
	--if combatType == "UNIT_DIED" then
	local timestamp, combatType, something, srcGUID, srcName, srcFlags, something2, destGUID, destName, destFlags = unpack(args)
	if combatType == "PARTY_KILL" then
		if not isMythicPlus() then return end
		if mppDebug then
			--foreach(args, print)
		end
		local npcID = getNPCID(destGUID)
		if npcID then
			local isDataUseful = true
			local timeSinceLastKill = GetTime() - lastKill[1]
			if timeSinceLastKill <= getSetting("inconclusiveDataThreshold") then
				debugLog("Data not useful: " .. timeSinceLastKill .. " - " .. lastKill[1] .. " - " .. GetTime())
				isDataUseful = false
			end
			lastKill = {GetTime(), npcID, destName, isDataUseful} -- timestamp is not at all accurate, we use GetTime() instead.
			if mppDebug then
				foreach(lastKill, print)
			end
		end
	end
end

local function verifyDB(forceWipe)
	if not MythicPlusProgressDB or not MythicPlusProgressDB["settings"] or not MythicPlusProgressDB["npcData"] or forceWipe then
		print("Running first time MPP setup. This should only happen once. Enjoy! ;)")
		MythicPlusProgressDB = {}
		MythicPlusProgressDB["settings"] = defaultSettings
		MythicPlusProgressDB["npcData"] = {}
		if mppDefaultValues ~= nil then
			for k,v in pairs(mppDefaultValues) do
				npcID, value, name = unpack(v)
				updateValue(npcID, value, name, 3)
			end
		end
	end
	-- Ensure we have all settings
	for setting, value in pairs(defaultSettings) do
		if MythicPlusProgressDB["settings"][setting] == nil then
			MythicPlusProgressDB["settings"][setting] = value
		end
	end
end

-- Called directly by our hook
local function onAddonLoad(addonName)
	if addonName == "MythicPlusProgress" then
		verifyDB()
		if isMythicPlus() then
			quantity = getEnemyForcesProgress()
			debugLog("MPP Loaded in progress: " .. quantity .. "in.")
		else
			quantity = 0
			debugLog("MPP loaded not in progress.")
		end
		loaded = true
	end
end

---
--- TOOLTIPS
---
	
local function addLineToTooltip(str)
    GameTooltip:AddDoubleLine(str)
    GameTooltip:Show()
end

local function shouldAddTooltip(unit)
	if loaded and getSetting("enabled") and getSetting("enableTooltip") and isMythicPlus() and isValidTarget(unit) then
		return true
	end
	return false
end

-- TODO: Color after certainty?
local function getTooltipMessage(npcID)
	local tempMessage = "|cFF"..getSetting("tooltipColor").."M+Progress: "
	local estProg = getEstimatedProgress(npcID)
	if not estProg then
		return tempMessage .. "No record."
	end
	mobsLeft = (100 - getEnemyForcesProgress()) / estProg
	tempMessage = string.format("%s%.2f%s (%i left)", tempMessage, estProg, "%", math.ceil(mobsLeft))
	return tempMessage
end
	
local function onNPCTooltip(self)
	local unit = select(2, self:GetUnit())
	if unit then
		local guid = UnitGUID(unit)
		npcID = getNPCID(guid)
		if npcID and shouldAddTooltip(unit) then
			local tooltipMessage = getTooltipMessage(npcID)
			if tooltipMessage then
				addLineToTooltip(tooltipMessage)
			end
		end
	end
end

---
--- SHITTY CURRENT PULL FRAME
---

currentPullFrame = CreateFrame("frame", "currentPullFrame12", UIParent)
mppFrame = currentPullFrame
currentPullFrame:SetPoint("CENTER", UIParent, 400, 300)
currentPullFrame:EnableMouse(true)
currentPullFrame:SetMovable(true)
currentPullFrame:RegisterForDrag("LeftButton")
currentPullFrame:SetScript("OnDragStart", currentPullFrame.StartMoving)
currentPullFrame:SetScript("OnDragStop", currentPullFrame.StopMovingOrSizing)
currentPullFrame:SetWidth(50)
currentPullFrame:SetHeight(50)
--currentPullFrame:SetAllPoints()
--currentPullFrame:SetClampRectInsets(200, 400, 200, 500)
currentPullText = currentPullFrame:CreateFontString("currentPullString", "BACKGROUND", "GameFontHighlightLarge")
currentPullText:SetPoint("CENTER");
currentPullText:SetText("MPP String Uninitialized.")

---
--- NAMEPLATES
---

local function isTargetPulled(target)
	-- debugLog("isTargetPulled with args: " ..target)
	local threat = UnitThreatSituation("player", target) or -1 -- Is nil if we're not on their aggro table, so make it -1 instead.
	if isValidTarget(target) and (threat >= 0 or UnitPlayerControlled(target.."target")) then
		return true
	end
	return false
end
	
local function getPulledUnits()
	local tempList = {}
	for token, guid in pairs(activeNameplates) do
		if isTargetPulled(token) then
			table.insert(tempList, guid)
		end
	end
	return tempList
end

local function getPulledProgress(pulledUnits)
	local estProg = 0
	for _, guid in pairs(pulledUnits) do
		npcID = getNPCID(guid)
		if npcID then
			estProg = estProg + (getEstimatedProgress(npcID) or 0)
		end
	end
	return estProg
end

local function shouldShowNameplateEstimate()
	if getSetting("enabled") and getSetting("enablePullEstimate") and isMythicPlus() and not isDungeonFinished() then
		return true
	end
	return false
end

local function setEstimateLabel(s)
	currentPullString:SetText(s)
	currentPullFrame:SetWidth(currentPullString:GetStringWidth())
	currentPullFrame:SetHeight(currentPullString:GetStringHeight())
	--print(currentPullFrame:GetCenter())
end

local function updateNameplateEstimate()
	if not shouldShowNameplateEstimate() then
		currentPullFrame:Hide()
		return
	else
		currentPullFrame:Show()
	end
	local pulledUnits = getPulledUnits()
	local estProg = getPulledProgress(pulledUnits)
	-- debugLog(tlen(pulledUnits) .. "/" .. tlen(activeNameplates).." active nameplates for an estimated " .. estProg .. "%")
	local curProg = getEnemyForcesProgress()
	local totProg = estProg + curProg
	if estProg == 0 then
		tempMessage = "No recorded mobs pulled or nameplates inactive."
	else
		tempMessage = string.format("Current pull: %.2f%s + %.2f%s = %.2f%s", curProg, "%", estProg, "%", (math.floor(totProg*100)/100), "%")
	end
	setEstimateLabel(tempMessage)
end

local function onAddNameplate(token)
	local tempGUID = UnitGUID(token)
	if tempGUID then
		activeNameplates[token] = tempGUID
	end
end

local function onRemoveNameplate(token)
	activeNameplates[token] = nil
end

---
--- SET UP HOOKS
---

local frame = CreateFrame("FRAME")
frame:RegisterEvent("ADDON_LOADED")
frame:RegisterEvent("SCENARIO_CRITERIA_UPDATE")
frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")


frame:RegisterEvent("NAME_PLATE_CREATED")
frame:RegisterEvent("NAME_PLATE_UNIT_ADDED")
frame:RegisterEvent("NAME_PLATE_UNIT_REMOVED")

function frame:OnEvent(event, ...)
	args={...}
	if event == "ADDON_LOADED" then
		onAddonLoad(args[1])
	elseif event == "SCENARIO_CRITERIA_UPDATE" then
		onCriteriaUpdate()
	elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
		onCombatLogEvent(args)
	elseif event == "NAME_PLATE_CREATED" then
		--onCreateNameplate(...)
	elseif event == "NAME_PLATE_UNIT_ADDED" then
		onAddNameplate(...)
	elseif event == "NAME_PLATE_UNIT_REMOVED" then
		onRemoveNameplate(...)
	else
		print("MythicPlusProgress unhandled event: " .. event)
		for k,v in ipairs(args) do
			print(k,v)
		end
	end
end

function frame:OnUpdate(elapsed)
	nameplateUpdateTimer = nameplateUpdateTimer + elapsed * 1000 -- Not using milliseconds in 2016? WutFace
	if nameplateUpdateTimer >= getSetting("nameplateUpdateRate") then
		nameplateUpdateTimer = 0
		updateNameplateEstimate()
	end
end

frame:SetScript("OnEvent", frame.OnEvent)
frame:SetScript("OnUpdate", frame.OnUpdate)
GameTooltip:HookScript("OnTooltipSetUnit", onNPCTooltip)

SLASH_MYTHICPLUSPROGRESS1,SLASH_MYTHICPLUSPROGRESS2,SLASH_MYTHICPLUSPROGRESS3 = "/mpp","/mypp","/mythicplusprogress"

-- TODO: Redo this entire thing that I threw together in 2 min
function SlashCmdList.MYTHICPLUSPROGRESS(args, editbox)
	args = split(string.lower(args))
	if args[1] == "toggle" then
		print("MythicPlusProgress toggled ".. (toggleSetting("enabled") and "ON" or "OFF"))
	elseif args[1] == "reset" then
		currentPullFrame:SetPoint("CENTER", UIParent, 50, 50)
		print("MythicPlusProgress' position reset.")
	elseif args[1] == "sim" then
		mppSimulationMode = not mppSimulationMode
		print("MPP simulation mode toggled ".. (mppSimulationMode and "ON" or "OFF"))
	elseif args[1] == "debug" then
		mppDebug = not mppDebug
		print("MPP debug toggled ".. (mppSimulationMode and "ON" or "OFF"))
	elseif args[1] == "updatevalue" then
		updateValue(getNPCID(UnitGUID("target")), args[2])
	elseif args[1] == "simmax" then
		simulationMax = args[2]
	elseif args[1] == "simcur" then
		simulationCurrent = args[2]
	elseif args[1] == "wipeall" then
		verifyDB(true)
		print("RIP Database.")
	elseif args[1] == "tooltip" then
		setSetting("enableTooltip", not getSetting("enableTooltip"))
		print("Toggled tooltips ".. (getSetting("enableTooltip") and "on" or "off"))
	else
		print("MythicPlusProgress commands:")
		print("/mpp toggle (Toggles addon visibility on/off, it will still record npc data while in Mythic+)")
		print("/mpp reset (reset position of 'current pull' text, it likes to run away.)")
		print("/mpp tooltip (toggles the tooltip functionality only on/off)")
	end
end